unit ImgHolder;

interface

uses
    Windows, SysUtils, Classes, Graphics, Math, Affine;

//=============================================================================
//  摜ƃev[g摜ێNX
//=============================================================================
type
    TImageHolder = class(TObject)
    protected

        //  摜TCY
        FImageW: Integer;
        FImageH: Integer;

        //  ev[gTCY
        FTemplateW: Integer;
        FTemplateH: Integer;

        //  Ɨp
        FWork: TBitmap;

        //  摜ێ
        FImage: TBitmap;

        //  ev[g摜ێ
        FTemplates: TList;

        //  O[XP[̃pbg쐬
        function BuildPallete(): HPalette;

        //  O[XP[ϊ
        procedure GrayScaleByDriver(const Source: TBitmap; Target: TBitmap);
        procedure GrayScaleByNTSC(const Source: TBitmap; Target: TBitmap);
        procedure GrayScaleByAverage(const Source: TBitmap; Target: TBitmap);
        procedure GrayScaleBySimple(const Source: TBitmap; Target: TBitmap);

        //  `tB^
        procedure LinearSmoothing(Target: TBitmap);

        //  摜Zbg
        procedure SetImage(const Source: TBitmap);

        //  ev[g摜擾
        function GetTemplate(AIndex: Integer): TBitmap;

    public

        //  vpeB
        property Image: TBitmap read FImage write SetImage;
        property Template[Index: Integer]: TBitmap read GetTemplate;

    public
        //  ev[g摜ǉ
        procedure AddTemplate(const Source: TBitmap);

        //  ev[g摜NA
        procedure ClearTemplate();

        //  摜𕽊
        procedure SmoothingImage();

        //  ev[g摜𕽊
        procedure SmoothingTemplate();

        //  RXgN^
        constructor Create();

        //  fXgN^
        destructor Destroy(); override;
    end;

//=============================================================================
//  w肳ꂽp[^Ō摜ƃev[g摜̎ĂxvZNX
//=============================================================================
type
    TEvalFitness = class(TImageHolder)
    protected

        //  ]vZNX
        FRotater: TRotatePoints;

        //  ]p摜
        FEvalBuf: TBitmap;

    public
        //  ]
        function Eval(const TemplateIndex: Integer; const AX, AY, ARotate, ARate: Double): Double;

        //  RXgN^
        constructor Create();

        //  fXgN^
        destructor Destroy(); override;

    public
        property EvalBuf: TBitmap read FEvalBuf;
        property Rotater: TRotatePoints read FRotater;
    end;

var
    GImageHolder: TEvalFitness = nil;

implementation

//=============================================================================
//  摜ƃev[g摜ێNX
//=============================================================================

//-----------------------------------------------------------------------------
//  RXgN^
constructor TImageHolder.Create();
begin
    inherited;

    //  Ɨp
    FWork  := TBitmap.Create();

    //  摜쐬
    FImage := TBitmap.Create();

    //  摜ɃpbgKp
    FImage.PixelFormat := pf8bit;
    FImage.Palette := BuildPallete();

    //  ev[g摜Xg쐬
    FTemplates := TList.Create();
end;

//-----------------------------------------------------------------------------
//  fXgN^
destructor TImageHolder.Destroy();
begin
    //  Ɨprbg}bvj
    FWork.Free();

    //  摜j
    DeleteObject(FImage.Palette);
    FImage.Free();

    //  ev[g摜Xgj
    ClearTemplate();
    FTemplates.Free();

    inherited;
end;

//-----------------------------------------------------------------------------
//  O[XP[̃pbg쐬
function TImageHolder.BuildPallete(): HPalette;
var
    i: Integer;
    LogPalette: TMaxLogPalette;
begin
    LogPalette.palNumEntries := 256;
    LogPalette.palVersion := $0300;
    for i := 0 to 255 do
    begin
        LogPalette.palPalEntry[i].peRed   := 255 - i;
        LogPalette.palPalEntry[i].peGreen := 255 - i;
        LogPalette.palPalEntry[i].peBlue  := 255 - i;
    end;
    Result := CreatePalette(PLogPalette(@LogPalette)^);
end;

//-----------------------------------------------------------------------------
//  Subject: [Delphi:48275] Re: ͂߂܂
type
    TTriple = packed record B, G, R: Byte; end;
    PTriple = ^TTriple;

//-----------------------------------------------------------------------------
//  O[XP[ϊ
//  rfIhCoɗ(^^;@ԑ
procedure TImageHolder.GrayScaleByDriver(const Source: TBitmap; Target: TBitmap);
begin
    //  Ĥ߂ɃC[W]L
    FWork.Width := Source.Width;
    FWork.Height := Source.Height;
    FWork.PixelFormat := pf24bit;
    BitBlt(FWork.Canvas.Handle, 0, 0, FWork.Width, FWork.Height,
           Source.Canvas.Handle, 0, 0, SRCCOPY);

    //  ϊ
    Target.Canvas.Draw(0, 0, FWork);
end;

//-----------------------------------------------------------------------------
//  O[XP[ϊ
//  NTSC dϖ@@łiǂ
procedure TImageHolder.GrayScaleByNTSC(const Source: TBitmap; Target: TBitmap);
var
    x, y: Integer;
    pSource: PTriple;
    pDest: PByte;
begin
    //  Ĥ߂ɃC[W]L
    FWork.Width := Source.Width;
    FWork.Height := Source.Height;
    FWork.PixelFormat := pf24bit;
    BitBlt(FWork.Canvas.Handle, 0, 0, FWork.Width, FWork.Height,
           Source.Canvas.Handle, 0, 0, SRCCOPY);

    //  ϊ
    for y := 0 to FWork.Height - 1 do
    begin
        pSource := FWork.ScanLine[y];
        pDest := Target.ScanLine[y];

        for x := 0 to FWork.Width - 1 do
        begin
            pDest^ := Round(255 - pSource.R * 0.298912 -
                                  pSource.G * 0.586611 -
                                  pSource.B * 0.114478);
            Inc(pSource);
            Inc(pDest);
        end;
    end;
end;

//-----------------------------------------------------------------------------
//  O[XP[ϊ
//  Ԓl@@Ȃ񂩂̂؂Ƃ摜ɂȂ݂łB܂悭Ȃ
procedure TImageHolder.GrayScaleByAverage(const Source: TBitmap; Target: TBitmap);
var
    x, y: Integer;
    pSource: PTriple;
    pDest: PByte;
    Max,Min: Byte;
begin
    //  Ĥ߂ɃC[W]L
    FWork.Width := Source.Width;
    FWork.Height := Source.Height;
    FWork.PixelFormat := pf24bit;
    BitBlt(FWork.Canvas.Handle, 0, 0, FWork.Width, FWork.Height,
           Source.Canvas.Handle, 0, 0, SRCCOPY);

    //  ϊ
    for y := 0 to FWork.Height - 1 do
    begin
        pSource := FWork.ScanLine[y];
        pDest := Target.ScanLine[y];

        for x := 0 to FWork.Width - 1 do
        begin
            Max := pSource^.R;
            Min := pSource^.R;
            if pSource^.G > Max then Max := pSource^.G;
            if pSource^.B > Max then Max := pSource^.B;
            if pSource^.G < Min then Min := pSource^.G;
            if pSource^.B < Min then Min := pSource^.B;
            pDest^ := 255 - (Max + Min) div 2;
            Inc(pSource);
            Inc(pDest);
        end;
    end;
end;

//-----------------------------------------------------------------------------
//  O[XP[ϊ
//  Pϖ@@@邭Ȃ肷܂ˁBƂ悭Ȃ
procedure TImageHolder.GrayScaleBySimple(const Source: TBitmap; Target: TBitmap);
var
    x, y: Integer;
    pSource: PTriple;
    pDest: PByte;
begin
    //  Ĥ߂ɃC[W]L
    FWork.Width := Source.Width;
    FWork.Height := Source.Height;
    FWork.PixelFormat := pf24bit;
    BitBlt(FWork.Canvas.Handle, 0, 0, FWork.Width, FWork.Height,
           Source.Canvas.Handle, 0, 0, SRCCOPY);

    //  ϊ
    for y := 0 to FWork.Height - 1 do
    begin
        pSource := FWork.ScanLine[y];
        pDest := Target.ScanLine[y];

        for x := 0 to FWork.Width - 1 do
        begin
            pDest^ := 255 - (pSource.R + pSource.G + pSource.B) div 3;
            Inc(pSource);
            Inc(pDest);
        end;
    end;
end;

//-----------------------------------------------------------------------------
//  `tB^
procedure TImageHolder.LinearSmoothing(Target: TBitmap);
const
    WEIGHT: array[0..2] of array[0..2] of Byte
            = ((1, 1, 1), (1, 2, 1), (1, 1, 1));
    DIVVAL = 10;
var
    x, y: Integer;
    pSource: array[0..2] of PByteArray;
    pDest: PByteArray;
begin
    //  Ĥ߂ɃC[W]L
    FWork.Width := Target.Width;
    FWork.Height := Target.Height;
    FWork.PixelFormat := pf8bit;
    BitBlt(FWork.Canvas.Handle, 0, 0, FWork.Width, FWork.Height,
           Target.Canvas.Handle, 0, 0, SRCCOPY);

    //  摜̒[1sNZ͕u(Ƃ肠)
    for y := 1 to FWork.Height - 2 do
    begin
        pSource[0] := FWork.ScanLine[y - 1];
        pSource[1] := FWork.ScanLine[y];
        pSource[2] := FWork.ScanLine[y + 1];
        pDest := Target.ScanLine[y];

        for x := 1 to FWork.Width - 2 do
        begin
            pDest[x] := Trunc((  WEIGHT[0][0] * pSource[0][x - 1]
                               + WEIGHT[0][1] * pSource[0][x]
                               + WEIGHT[0][2] * pSource[0][x + 1]
                               + WEIGHT[1][0] * pSource[1][x - 1]
                               + WEIGHT[1][1] * pSource[1][x]
                               + WEIGHT[1][2] * pSource[1][x + 1]
                               + WEIGHT[2][0] * pSource[2][x - 1]
                               + WEIGHT[2][1] * pSource[2][x]
                               + WEIGHT[2][2] * pSource[2][x + 1]) / DIVVAL);
        end;
    end;
end;

//-----------------------------------------------------------------------------
//  摜Zbg
procedure TImageHolder.SetImage(const Source: TBitmap);
begin
    //  E͕ʂɂێ
    FImageW := Source.Width;
    FImageH := Source.Height;

    //  ŏIC[Wݒ
    FImage.Width := FImageW;
    FImage.Height := FImageH;

    //  ϊ
    if (Source.PixelFormat <> pf8bit) then
    begin
        GrayScaleByDriver(Source, FImage);
        //GrayScaleByNTSC(Source, FImage);
        //GrayScaleByAverage(Source, FImage);
        //GrayScaleBySimple(Source, FImage);
    end
    else
    begin
        FImage.Canvas.Draw(0, 0, Source);
    end;
end;

//-----------------------------------------------------------------------------
//  ev[g摜擾
function TImageHolder.GetTemplate(AIndex: Integer): TBitmap;
begin
    //  ͈̓`FbNT{
    Result := TBitmap(FTemplates.Items[AIndex]);
end;

//-----------------------------------------------------------------------------
//  ev[g摜ǉ
procedure TImageHolder.AddTemplate(const Source: TBitmap);
var
    Template: TBitmap;
begin
    //  ŏ̂ЂƂŃTCY肷
    if (FTemplates.Count = 0) then
    begin
        FTemplateW := Source.Width;
        FTemplateH := Source.Height;
    end;

    //  ́ÃTCYȊO͎󂯕tȂ
    if (FTemplateW = Source.Width) and (FTemplateH = Source.Height) then
    begin
        //  Vev[g쐬
        Template := TBitmap.Create();

        //  ݒ
        Template.Width := FTemplateW;
        Template.Height := FTemplateH;
        Template.PixelFormat := pf8bit;
        Template.Palette := BuildPallete();

        //  ϊ
        if (Source.PixelFormat <> pf8bit) then
        begin
            GrayScaleByDriver(Source, Template);
            //GrayScaleByNTSC(Source, Template);
            //GrayScaleByAverage(Source, Template);
            //GrayScaleBySimple(Source, Template);
        end
        else
        begin
            Template.Canvas.Draw(0, 0, Source);
        end;

        //  Xgɒǉ
        FTemplates.Add(Template);
    end;
end;

//-----------------------------------------------------------------------------
//  ev[g摜NA
procedure TImageHolder.ClearTemplate();
begin
    while (FTemplates.Count > 0) do
    begin
        DeleteObject(Template[0].Palette);
        Template[0].Free();
        FTemplates.Delete(0);
    end;
end;

//-----------------------------------------------------------------------------
//  摜𕽊
procedure TImageHolder.SmoothingImage();
begin
    LinearSmoothing(FImage);
end;

//-----------------------------------------------------------------------------
//  ev[g摜𕽊
procedure TImageHolder.SmoothingTemplate();
var
    i: Integer;
begin
    for i := 0 to FTemplates.Count - 1 do
    begin
        LinearSmoothing(Template[i]);
    end;
end;

//=============================================================================
//  w肳ꂽp[^Ō摜ƃev[g摜̎ĂxvZNX
//=============================================================================


//-----------------------------------------------------------------------------
//  RXgN^
constructor TEvalFitness.Create();
begin
    inherited;
    FRotater := TRotatePoints.Create();
    FEvalBuf := TBitmap.Create();
    FEvalBuf.PixelFormat := pf8bit;
    FEvalBuf.Palette := BuildPallete();
end;

//-----------------------------------------------------------------------------
//  fXgN^
destructor TEvalFitness.Destroy();
begin
    FRotater.Free();
    DeleteObject(FEvalBuf.Palette);
    FEvalBuf.Free();
    inherited;
end;

//-----------------------------------------------------------------------------
//  ]
function TEvalFitness.Eval(const TemplateIndex: Integer; const AX, AY, ARotate, ARate: Double): Double;
var
    Temp: TBitmap;
    x, y: Integer;
    r: TRect;
    PtEx: TPointEx;
    Pt: TPoint;
    DiffX, DiffY: Double;
    TL, TR, BL, BR: Byte;
    TValue, BValue: Double;
    IsRight, IsBottom: Boolean;
    Sum, Diff: Int64;
begin
    //  ev[g摜o
    if (0 <= TemplateIndex) and (TemplateIndex < FTemplates.Count) then
    begin
        Temp := GetTemplate(TemplateIndex);
    end
    else
    begin
        Result := 0;
        Exit;
    end;

    //  ]vZNXɃp[^ݒ

    //  sړ()
    FRotater.TranslateX := AX;

    //  sړ(c)
    FRotater.TranslateY := AY;

    //  ]px
    FRotater.Radian := DegToRad(ARotate);

    //  g嗦
    FRotater.Rate   := ARate;

    //  ev[g摜̋`ݒ
    FRotater.SetPoint(Bounds(0, 0, FTemplateW, FTemplateH));

    //  ]̒Sݒ
    FRotater.SetCenterOfPts();

    //  `](̃NXł͕svBʒu`悵ȂƂ̂߂)
    FRotater.Execute();

    //  ]p摜
    FEvalBuf.Width := FTemplateW;
    FEvalBuf.Height := FTemplateH;
    for y := 0 to FTemplateH - 1 do
    begin
        FillChar(FEvalBuf.ScanLine[y]^, FTemplateW, 0);
    end;

    //  摜̗̈
    r := Bounds(0, 0, FImageW, FImageH);

    //  ev[g摜摜ɏdˍ킹͈̔͂̉摜
    //  ev[g摜Ɣrł悤ɁuĂv
    for y := 0 to FTemplateH - 1 do
    begin
        for x := 0 to FTemplateW - 1 do
        begin
            //  ̓_ϊ_vZ
            PtEx := FRotater.Execute(PointEx(x, y));

            //  iq_̓_
            Pt := PointExToPoint(PtEx);

            //  ۂ̊iq_ƕϊ_Ƃ̍߂
            DiffX := PtEx.X - Pt.X; 
            DiffY := PtEx.Y - Pt.Y;

            //  摜̒ɓĂꍇ̂ݏ
            if (PtInRect(r, Pt)) then
            begin
                //  _Ԃiq_̐Fo

                //  (͂܂Ɋiq_̂)
                TL := PByteArray(FImage.ScanLine[Pt.y])[Pt.x];

                //  ̓_Eɓ_Ȃ
                if (Pt.X < FImageW - 1) then
                begin
                    //  E̓_ɂ
                    TR := PByteArray(FImage.ScanLine[Pt.y])[Pt.x + 1];
                    IsRight := False;
                end

                //  ȂȂ
                else
                begin
                    //  E̓_͍Ɉv
                    TR := TL;
                    IsRight := True;
                end;

                //  ̓_ɓ_Ȃ
                if (Pt.Y < FImageH - 1) then
                begin
                    //  ̓_ɂ
                    BL := PByteArray(FImage.ScanLine[Pt.y + 1])[Pt.x];
                    IsBottom := False;
                end

                //  ȂȂ
                else
                begin
                    //  ̓_͍Ɉv
                    BL := TL;
                    IsBottom := True;
                end;

                //  ̓_E[ł[łȂȂ
                if (not IsRight) and (not IsBottom) then
                begin
                    //  E̓_ɂ
                    BR := PByteArray(FImage.ScanLine[Pt.y + 1])[Pt.x + 1];
                end

                //  ȊȌꍇ
                else
                begin
                    //  E̓_͍̓_Ɉv
                    BR := BL;
                end;

                //  TL(TopLeft) TR(TopRight) ̒Ԓl
                TValue := TL * (1 - DiffX) + TR * DiffX;

                //  TL(BottomLeft) TR(BottomRight) ̒Ԓl
                BValue := BL * (1 - DiffX) + BR * DiffX;

                //  ̒Ԓl𐶐_̒lƂ
//                PByteArray(FEvalBuf.ScanLine[y])[x] := Round(255 - (TValue * (1 - DiffY) + BValue * DiffY));
                PByteArray(FEvalBuf.ScanLine[y])[x] := Round((TValue * (1 - DiffY) + BValue * DiffY));
            end;
        end;
    end;

    //  ĂxvZ
    Sum := 0;
    for y := 0 to FTemplateH - 1 do
    begin
        for x := 0 to FTemplateW - 1 do
        begin
            //  ̉摜ƃev[g摜̊Kl̍
            Diff := Abs(PByteArray(FEvalBuf.ScanLine[y])[x] - PByteArray(Temp.ScanLine[y])[x]);

            //  ̉摜ƃev[g摜̊Kl̍𑫂ł
            Inc(Sum, Diff);

            //  ̉摜ƃev[g摜̊Kl̍2𑫂ł
            //Inc(Sum, Diff * Diff);
        end;
    end;

    //  uĂxv 0..1 ɕϊ(1 ɋ߂قǎĂ)
    Diff := 255 * FTemplateH * FTemplateW;
    Result := 1 - Sum / Diff;
    //Result := Sum / (Diff * Diff);
end;

//-----------------------------------------------------------------------------
initialization
    GImageHolder := TEvalFitness.Create();
finalization
    GImageHolder.Free();
end.
